import logging
from pathlib import Path
from typing import BinaryIO, Dict, Mapping

import requests

from kognic.base_clients.cloud_storage.upload_spec import UploadSpec, UploadableData
from kognic.base_clients.retry_support import request_with_retry

log = logging.getLogger(__name__)


class UploadHandler:

    def __init__(self, max_retry_attempts: int = 23, max_retry_wait_time: int = 60, timeout: int = 60) -> None:
        """
        :param max_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_retry_wait_time:  Max with time before retrying an upload to GCS.
        :param timeout: Max time to wait for response from server.
        """
        self.max_num_retries = max_retry_attempts
        self.max_retry_wait_time = max_retry_wait_time  # seconds
        self.timeout = timeout  # seconds

    #  Using similar retry strategy as gsutil
    #  https://cloud.google.com/storage/docs/gsutil/addlhelp/RetryHandlingStrategy
    def _upload_file(self, upload_url: str, file: UploadableData, headers: Dict[str, str]) -> None:
        """
        Upload the file to GCS, retries if the upload fails with some specific status codes or timeouts.
        """
        request_with_retry(requests.put,
                           self.max_num_retries,
                           self.max_retry_wait_time,
                           url=upload_url,
                           data=file,
                           headers=headers)

    def upload_files(self, upload_specs: Mapping[str, UploadSpec]) -> None:
        """
        Upload all files to cloud storage

        :param upload_specs: map between filename and details of what to upload where
        """
        for (resource_id, upload_spec) in upload_specs.items():
            if upload_spec.data is None and upload_spec.callback is None:
                self._upload_from_local_file(upload_spec)
            elif upload_spec.data is not None:
                self._upload_from_blob(upload_spec)
            elif upload_spec.callback is not None:
                self._upload_from_callback(upload_spec)

    def upload_file(self, file: BinaryIO, url: str) -> None:
        headers = {"Content-Type": "application/json"}
        self._upload_file(url, file, headers)

    def _upload(self, upload_spec: UploadSpec, data: UploadableData):
        headers = {"Content-Type": upload_spec.content_type}
        self._upload_file(upload_spec.destination, data, headers)

    def _upload_from_blob(self, upload_spec: UploadSpec):
        log.debug(f"Blob upload for filename={upload_spec.filename}")
        self._upload(upload_spec, upload_spec.data)

    def _upload_from_local_file(self, upload_spec: UploadSpec):
        log.debug(f"Upload from local file for filename={upload_spec.filename}")
        file = Path(upload_spec.filename).expanduser().open('rb')
        self._upload(upload_spec, file)

    def _upload_from_callback(self, upload_spec: UploadSpec):
        log.debug(f"Upload from callback for filename={upload_spec.filename}")
        try:
            data = upload_spec.callback(upload_spec.filename)
            self._upload(upload_spec, data)
        except Exception as e:
            raise Exception("Failed to upload file: callback failed", e)
        pass



from typing import Union, Optional, Dict

from humps import camelize
from pydantic import BaseModel

CursorIdType = Union[int, str]


def to_camel_case(string: str) -> str:
    return camelize(string)


class BaseSerializer(BaseModel):

    @classmethod
    def from_json(cls, js: Dict):
        return cls.parse_obj(js)

    def to_dict(self, by_alias=True) -> Dict:
        return self.dict(exclude_none=True, by_alias=by_alias)

    class Config:
        alias_generator = to_camel_case
        orm_mode = True
        allow_population_by_field_name = True


class PageMetadata(BaseSerializer):
    next_cursor_id: Optional[CursorIdType]


class PaginatedResponse(BaseSerializer):
    data: Union[dict, list]
    metadata: PageMetadata



from io import IOBase
from typing import Optional, Callable, Union, BinaryIO

UploadableData = Union[bytes, BinaryIO, IOBase]


class UploadSpec:
    """
    Represents an upload to be performed to some resource ID: what is to be uploaded and where.
    A filename must always be set, and a data blob or callback may optionally be set. With just a filename this upload
    will use the named file. With a filename and a blob, the data blob is used and the filename is informational only.
    If a callback is set, it will be called with the filename as its argument and must return a blob (bytes) for upload.
    """
    destination: str
    filename: str
    content_type: str
    data: Optional[UploadableData]
    callback: Optional[Callable[[str], UploadableData]]

    def __init__(self, destination: str, content_type: str, filename: str = None, data: Optional[UploadableData] = None,
                 callback: Optional[Callable[[str], UploadableData]] = None):
        self.destination = destination
        self.filename = filename
        self.content_type = content_type
        self.data = data
        self.callback = callback

    def __repr__(self):
        data_part = '' if self.data is None else f', data'
        cb_part = '' if self.callback is None else f', callback'
        return f"UploadSpec({self.content_type}, destination={self.destination}, filename={self.filename}{data_part}{cb_part})"



"""Client for communicating with the Kognic platform."""
import logging
from pathlib import Path
from typing import Dict, Mapping, Optional, BinaryIO

from kognic.base_clients.cloud_storage.download_handler import DownloadHandler
from kognic.base_clients.cloud_storage.upload_handler import UploadHandler
from kognic.base_clients.cloud_storage.upload_spec import UploadSpec

log = logging.getLogger(__name__)


class FileResourceClient:

    def __init__(self, max_retry_attempts: int = 23, max_retry_wait_time: int = 60, timeout: int = 60):
        """
        :param max_upload_retry_attempts: Max number of attempts to retry uploading a file to GCS.
        :param max_upload_retry_wait_time:  Max with time before retrying an upload to GCS.
        :param timeout: Max time to wait for response from server.
        """
        self._upload_handler = UploadHandler(max_retry_attempts, max_retry_wait_time, timeout)
        self._download_handler = DownloadHandler(max_retry_attempts, max_retry_wait_time, timeout)

    def upload_files(self, upload_specs: Mapping[str, UploadSpec]) -> None:
        """
        Upload all files to cloud storage

        :param upload_specs: Mapping from a source name to a destination and source of content for an upload
        """
        self._upload_handler.upload_files(upload_specs=upload_specs)

    def upload_json(self, file: BinaryIO, url: str) -> None:
        """
        Upload a single file to storage, using the specified url
        :param file: A binary representation of the file
        :param url: The url ot upload to file to
        """

        self._upload_handler.upload_file(file, url)

    def get_json(self, url: str) -> Dict:
        """
        Downloads a json from cloud storage

        :param url: Signed URL to GCS resource to download
        """
        return self._download_handler.get_json(url)
